#ifndef DMUTILS_GRIDFTP_C

#define _GNU_SOURCE
#include <errno.h>
#include "gridftp-ifce.h"
#include "globus_libc.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>


typedef struct {
    globus_mutex_t mutex;
    globus_cond_t cond;
    volatile globus_bool_t done;
    volatile globus_bool_t errflag;
    globus_object_t *error;
} monitor_t;

typedef struct {
    LONG64 transferred_bytes;
    float instant_throughput;
    float avg_throughput;
} performance;

typedef void (* copyfile_callback_t)(
        performance s_perf,
        struct timespec ts,
        int *timeoutflg
        );

typedef struct {
    monitor_t *mt;
    copyfile_callback_t *user_cb;
    volatile globus_bool_t done;
} check_t;

typedef struct {
    globus_bool_t restart;
    globus_ftp_client_restart_marker_t saved_marker;
} marker_store;

marker_store  restart_store;


#ifdef GLOBUS_ERROR_TYPE_MULTIPLE
#define GLOBUS_ERROR2STRING(errorobj) globus_error_print_friendly (errorobj)
#else
#define GLOBUS_ERROR2STRING(errorobj) globus_object_printable_to_string (errorobj)
#endif

int dmutils_gridftp_check_gsi_url(const char* url) {
        return strncmp(url, "gsiftp:", 7) == 0 ? 0 : -1;
    }


/*global variables section*/
static int internal_activation = 0;
static volatile int activated = 0;
 globus_gass_copy_handle_t ggc_handle;
/* the need for 2 is due to the fact that gass copy is
 * not copying attributes
 */
 globus_ftp_client_operationattr_t source_ftp_attr;
 globus_ftp_client_operationattr_t dest_ftp_attr;

 globus_gass_transfer_requestattr_t source_gass_attr;
 globus_gass_transfer_requestattr_t dest_gass_attr;

 globus_gass_copy_attr_t source_gass_copy_attr;
 globus_gass_copy_attr_t dest_gass_copy_attr;

  globus_ftp_client_handle_t ftp_handle;

 globus_ftp_client_handleattr_t ftp_handleattr;
 globus_gass_copy_handleattr_t ggc_handleattr;

 globus_ftp_client_operationattr_t ftp_attr;

 globus_ftp_client_plugin_t restart_plugin;
 globus_ftp_client_plugin_t restart_marker_plugin;

 globus_reltime_t interval;
 globus_abstime_t timeout_restart;
 globus_abstime_t * timeout_p = GLOBUS_NULL;

static int istimeout = 0;
static globus_bool_t globus_l_globus_url_copy_ctrlc = GLOBUS_FALSE;
static globus_bool_t globus_l_globus_url_copy_ctrlc_handled = GLOBUS_FALSE;

struct timespec usercb_ts;
performance s_transfer_performance;


/*
    The maximum number of times to retry the operation before giving
    up on the transfer.  Use 0 for infinite.  Default is 5"
 */
static int local_retries = 10;
/*
   The interval in seconds to wait after a failure before retrying"
   the transfer.  Use 0 for an exponential backoff.  Default is 0."
 */
static int local_interval = 5;

/*
    Maximum time after a failure to keep retrying.  Use 0 for no timeout.
 */
static int local_timeout = 10;

/**
 * Transfer begin callback
 * 
 * This callback is called when a get, put, or third party transfer is
 * started.
 *
 * The intended use for this callback is for the user to use the transfer
 * urls to locate a restart marker in some persistant storage. If one is
 * found, it should be copied into 'user_saved_marker' and the callback
 * should return GLOBUS_TRUE. This will cause the transfer to be restarted
 * using that restart marker.  If one is not found, return GLOBUS_FALSE to
 * indicate that the transfer should proceed from the beginning.
 *
 * In any case, this is also an opportunity for the user to set up any
 * storage in anticipation of restart markers for this transfer.
 */
globus_bool_t restart_marker_begin_cb(
        void * user_arg,
        globus_ftp_client_handle_t * handle,
        const char * source_url,
        const char * dest_url,
        globus_ftp_client_restart_marker_t * user_saved_marker) {

    marker_store * local_marker_store = (marker_store *) user_arg;
    if (local_marker_store && local_marker_store->restart == GLOBUS_TRUE) {
        globus_result_t result = globus_ftp_client_restart_marker_copy(user_saved_marker, &(local_marker_store->saved_marker));

        if (result == GLOBUS_SUCCESS) {           
            return GLOBUS_TRUE; /* user partial restart  / resume */
        }
    } else {
        return GLOBUS_FALSE;
    }

    return GLOBUS_FALSE;
}

/**
 * Restart marker received callback
 *
 * This callback will be called every time a restart marker is available.
 *
 * To receive restart markers in a 'put' or 'third_party_transfer', the
 * transfer must be in Extended Block mode. 'get' transfers will have their
 * markers generated internally.  Markers generated internally will be 'sent'
 * at most, once per second.
 *
 * The intended use for this callback is to allow the user to store this marker
 * (most likely in place of any previous marker) in a format that the
 * 'begin_cb' can parse and pass back.
 *
 */
void restart_marker_marker_cb(
        void * user_arg,
        globus_ftp_client_handle_t * handle,
        globus_ftp_client_restart_marker_t * marker) {

    marker_store * local_marker_store = (marker_store *) user_arg;
    if (local_marker_store && marker) {
        local_marker_store->restart = GLOBUS_TRUE;
        globus_ftp_client_restart_marker_destroy(&(local_marker_store->saved_marker));
        globus_ftp_client_restart_marker_copy(&(local_marker_store->saved_marker), marker);
    }
}

/**
 * Transfer complete callback
 *
 * This callback will be called upon transfer completion (successful or
 * otherwise)
 *
 */
void restart_marker_complete_cb(
        void * user_arg,
        globus_ftp_client_handle_t * handle,
        globus_object_t * error,
        const char * error_url) {

    if (error != GLOBUS_SUCCESS) {
        marker_store * local_marker_store = (marker_store *) user_arg;
        if (local_marker_store)
            local_marker_store->restart = GLOBUS_TRUE;
    } 
}

static pthread_mutex_t cs_mutex = PTHREAD_MUTEX_INITIALIZER;

void do_globus_activates(void) {
     pthread_mutex_lock( &cs_mutex );    
     if (activated == 0) 
        	activated = 1;
     else{
	        pthread_mutex_unlock( &cs_mutex );
     		return;
    }		
    pthread_mutex_unlock( &cs_mutex );


    //(void) globus_module_activate(GLOBUS_COMMON_MODULE);
    (void) globus_module_activate(GLOBUS_FTP_CLIENT_MODULE);
    (void) globus_module_activate(GLOBUS_GASS_COPY_MODULE);
    (void) globus_module_activate(GLOBUS_FTP_CLIENT_RESTART_PLUGIN_MODULE);
    //(void) globus_module_activate(GLOBUS_FTP_CLIENT_RESTART_MARKER_PLUGIN_MODULE);

    globus_result_t gresult;

    globus_gass_copy_attr_init(&source_gass_copy_attr);
    globus_gass_copy_attr_init(&dest_gass_copy_attr);
    /*
     * Intialize GridFTP handlers with the attributes specified
     */
    globus_gass_copy_handleattr_init(&ggc_handleattr);
    globus_ftp_client_handleattr_init(&ftp_handleattr);
    globus_ftp_client_handleattr_set_cache_all(&ftp_handleattr, (globus_bool_t) GLOBUS_TRUE);
    
    globus_bool_t gridftp2_getput = GLOBUS_FALSE;  
    globus_result_t gridftp2_enabled = globus_ftp_client_handleattr_get_gridftp2(&ftp_handleattr, &gridftp2_getput);
    if(gridftp2_enabled == GLOBUS_SUCCESS && gridftp2_getput == GLOBUS_FALSE){
    	globus_ftp_client_handleattr_set_gridftp2(&ftp_handleattr, GLOBUS_TRUE);    
    }
    
    /*
     * restart transfer works only for third party transfers because if one of the urls is
     * GLOBUS_GASS_COPY_URL_MODE_GASS then it returns GLOBUS_NULL to the restart callback handler
     */
    GlobusTimeReltimeSet(interval, local_interval, 0);

    if (local_timeout >= 0) {
        GlobusTimeAbstimeSet(timeout_restart, local_timeout, 0);
        timeout_p = &timeout_restart;
    }

    /* 
     * The restart plugin implements one scheme for providing reliability
     * functionality for the FTP Client library. 
     *
     * The specific functionality of this plugin is
     * to restart any FTP operation when a fault occurs. The plugin's operation
     * is parameterized to control how often and when to attempt to restart
     * the operation.
     *
     * This restart plugin will restart an FTP operation if a noticable
     * fault has occurred---a connection timing out, a failure by the server
     * to process a command, a protocol error, an authentication error.
     *
     * This plugin has three user-configurable parameters; these are
     * the maximum number of retries to attempt, the interval to wait between
     * retries, and the deadline after which no further retries will be attempted.
     * These are set by initializing a restart plugin instance with the function
     * globus_ftp_client_restart_plugin_init().
     *
     */
    gresult = globus_ftp_client_restart_plugin_init(
            &restart_plugin,
            local_retries,
            &interval,
            timeout_p);

    if (gresult != GLOBUS_SUCCESS) {
        globus_libc_fprintf(stderr, "Error: Unable to init restart_plugin plugin\n");
    }

    /*
     * This plugin is intended to allow users to make restart markers persistant.
     * In our case restart markers received back from the FTP server are kept in memory (no persistence)
     * During a transfer, every marker received will result in the user's 'marker'
     * callback being called with the new restart marker that can be stored. If
     * the application were to prematurely terminate (while transferring), the user
     * (after restarting the application) could pass this stored marker back to the
     * plugin via the 'begin' callback to force the transfer to be restarted from
     * the last marked point.
     */
    /*
    marker_store * restart_store = (marker_store *) globus_malloc(sizeof (marker_store));
    restart_store->restart = GLOBUS_FALSE;
    globus_ftp_client_restart_marker_init(&(restart_store->saved_marker));
    */
  /* 
    restart_store.restart = GLOBUS_FALSE;
    globus_ftp_client_restart_marker_init(&(restart_store).saved_marker);
 
    gresult = globus_ftp_client_restart_marker_plugin_init(
            &restart_marker_plugin,
            restart_marker_begin_cb,
            restart_marker_marker_cb,
            restart_marker_complete_cb,
            &restart_store);

    if (gresult != GLOBUS_SUCCESS) {
        globus_libc_fprintf(stderr, "Error: Unable to init restart_marker plugin\n");
    }

    gresult = globus_ftp_client_handleattr_add_plugin(&ftp_handleattr, &restart_marker_plugin);
    if (gresult != GLOBUS_SUCCESS) {
        globus_libc_fprintf(stderr, "Error: Unable to add restart_marker_plugin\n ");
    }
*/
    gresult = globus_ftp_client_handleattr_add_plugin(&ftp_handleattr, &restart_plugin);
    if (gresult != GLOBUS_SUCCESS) {
        globus_libc_fprintf(stderr, "Error: Unable to add restart_plugin\n ");
    }

    globus_ftp_client_operationattr_init(&ftp_attr);
    globus_gass_copy_handleattr_set_ftp_attr(&ggc_handleattr, &ftp_handleattr);
    globus_gass_copy_handle_init(&ggc_handle, &ggc_handleattr);

    ftp_handle = ggc_handle.ftp_handle;
}


void do_globus_deactivates(void) {
    pthread_mutex_lock( &cs_mutex );
    if (activated == 1)
        	activated = 0;
    else{
	    	pthread_mutex_unlock( &cs_mutex );
    		return;		
	}
    pthread_mutex_unlock( &cs_mutex );		

    globus_ftp_client_restart_plugin_destroy(&restart_plugin);
    // globus_ftp_client_restart_marker_plugin_destroy(&restart_marker_plugin);	
    globus_ftp_client_operationattr_destroy(&source_ftp_attr);
    globus_ftp_client_operationattr_destroy(&dest_ftp_attr);

    globus_gass_copy_handleattr_destroy(&ggc_handleattr);
    globus_ftp_client_handleattr_destroy(&ftp_handleattr);

    globus_ftp_client_operationattr_destroy(&ftp_attr);

    globus_gass_copy_handle_destroy(&ggc_handle);

    (void) globus_module_deactivate(GLOBUS_FTP_CLIENT_RESTART_PLUGIN_MODULE);
   // (void) globus_module_deactivate(GLOBUS_FTP_CLIENT_RESTART_MARKER_PLUGIN_MODULE);
    (void) globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
    (void) globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
    //(void) globus_module_deactivate(GLOBUS_COMMON_MODULE);
}

static int scan_errstring(const char *p) {

    int ret = EINVAL;
    if (p == NULL) return ret;

    if (strstr(p, "o such file"))
        ret = ENOENT;
    else if (strstr(p, "ermission denied") || strstr(p, "credential"))
        ret = EACCES;
    else if (strstr(p, "exists"))
        ret = EEXIST;
    return ret;
}

static void substr(char dest[], char src[], int offset, int len) {
    int i;
    for (i = 0; i < len && src[offset + i] != '\0'; i++)
        dest[i] = src[i + offset];
    dest[i] = '\0';
}

static void dmutils_gridftp_errmsg(
        char *errbuf,
        int errbufsz,
        const char *format,
        ...) {
    va_list ap;
    char *actual_format;

    if ((errbuf == NULL || format == NULL || format[0] == 0))
        return;

    va_start(ap, format);
    asprintf(&actual_format, "%s ", format);
    if (actual_format == NULL) return;

    if (errbuf == NULL)
        vfprintf(stderr, actual_format, ap);
    else
        vsnprintf(errbuf, errbufsz, actual_format, ap);

    free(actual_format);
}

/*
 *Interrupt handler to abort a transfer
 */
void globus_l_guc_interrupt_handler(
        void * user_arg) {
    monitor_t * monitor;
    monitor = (monitor_t *) user_arg;

    if(monitor){
    globus_mutex_lock(&monitor->mutex);
    {
        globus_l_globus_url_copy_ctrlc = GLOBUS_TRUE;
        globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->mutex);
    }
}

void dmutils_gridftp_gcallback(
        void* callback_arg,
        globus_ftp_client_handle_t* ftp_handle,
        globus_object_t* error) {
    monitor_t *monitor = (monitor_t *) callback_arg;
    
    if(monitor){
    globus_mutex_lock(&monitor->mutex);
    if (error != GLOBUS_SUCCESS) {
        monitor->errflag = GLOBUS_TRUE;
        monitor->error = globus_object_copy(error);
    }
    monitor->done = GLOBUS_TRUE;
    globus_cond_signal(&monitor->cond);
    globus_mutex_unlock(&monitor->mutex);
    }
}

void dmutils_gridftp_timeout_callback(void * user_arg) {
    check_t * check;
    monitor_t * monitor;
    copyfile_callback_t *cp_callback;
    globus_reltime_t delay;
    check = (check_t *) user_arg;
    monitor = (monitor_t *) check->mt;
    cp_callback = (copyfile_callback_t *) check->user_cb;
    globus_mutex_lock(&monitor->mutex);
    {
        GlobusTimeReltimeSet(delay, 1, 0);
        if (*cp_callback) {
            (*cp_callback)(s_transfer_performance, usercb_ts, &istimeout);
        }
        if (!monitor->done && !istimeout && !globus_l_globus_url_copy_ctrlc) {

            globus_callback_register_oneshot(
                    NULL,
                    &delay,
                    dmutils_gridftp_timeout_callback,
                    check);
        } else {
            check->done = GLOBUS_TRUE;
        }
    }
    globus_cond_signal(&monitor->cond);
    globus_mutex_unlock(&monitor->mutex);
}

globus_result_t dmutils_gridftp_client_wait(
        monitor_t *mp,
        globus_ftp_client_handle_t* gfhp,
        int timeout) {
    struct timespec ts;
    int save_errno = 0;
    
    if (timeout == -1)
        timeout = 0;


    if(mp && gfhp){
    globus_mutex_lock(&mp->mutex);

    if (timeout > 0) {
        ts.tv_sec = time(0) + timeout;
        ts.tv_nsec = 0;
        while (!mp->done && !save_errno) {
            save_errno = globus_cond_timedwait(&mp->cond, &mp->mutex, &ts);
            if (globus_l_globus_url_copy_ctrlc &&
                    !globus_l_globus_url_copy_ctrlc_handled) {
                globus_ftp_client_abort(gfhp);
                globus_l_globus_url_copy_ctrlc_handled = GLOBUS_TRUE;
            }


            if (save_errno == EINTR)
                save_errno = 0;
        }
    } else
        while (!mp->done && !save_errno) {
            save_errno = globus_cond_wait(&mp->cond, &mp->mutex);

            if (globus_l_globus_url_copy_ctrlc &&
                    !globus_l_globus_url_copy_ctrlc_handled) {
                globus_ftp_client_abort(gfhp);
                globus_l_globus_url_copy_ctrlc_handled = GLOBUS_TRUE;
            }


            if (save_errno == EINTR)
                save_errno = 0;
        }
    if (!mp->done) {
        int ret = 0;
        globus_ftp_client_abort(gfhp);
        do {
            ret = globus_cond_wait(&mp->cond, &mp->mutex);
            if (ret == EINTR)
                ret = 0;
        } while (!mp->done && !ret);
        if (ret && !save_errno)
            save_errno = ret;
    }

    globus_mutex_unlock(&mp->mutex);
    }

    if (save_errno) {
        errno = save_errno;
        return (-1);
    }

    return (0);
}

/*utility function	 */
globus_gass_copy_url_mode_t prepare_file_attr(
        const char* file_name,
        char* scope,
        int nbstreams,
        int tcp_bs,
        globus_gass_copy_attr_t * attr,
        globus_ftp_client_operationattr_t * oper_attr,
        globus_gass_transfer_requestattr_t* gass_attr) {

    /* Get Destination Url Mode */
    globus_gass_copy_url_mode_t url_mode;
    globus_ftp_control_tcpbuffer_t tcp_buffer;
    globus_ftp_control_dcau_t dcau;
    
    if(!file_name)
            return -1;

    /* Get Url Mode */
    globus_result_t gresult = globus_gass_copy_get_url_mode(file_name, &url_mode);
    if (GLOBUS_SUCCESS != gresult) {
        return -1;
    }

    /* switch on URL Mode */
    switch (url_mode) {
        case GLOBUS_GASS_COPY_URL_MODE_FTP:
        {
            /* Initialize FTP Attribute handles */
            globus_ftp_client_operationattr_init(oper_attr);
            globus_ftp_client_operationattr_set_resume_third_party_transfer(oper_attr, GLOBUS_TRUE);
            globus_ftp_control_parallelism_t parallelism __attribute__((unused));

            if (nbstreams > 1) {
  	        globus_ftp_client_operationattr_set_mode(oper_attr, GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
                parallelism.mode = GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
                parallelism.fixed.size = nbstreams;
                globus_ftp_client_operationattr_set_parallelism(oper_attr, &parallelism);
            }

            if (tcp_bs > 0) {
                /* set TCP bs */
                tcp_buffer.mode = GLOBUS_FTP_CONTROL_TCPBUFFER_FIXED;
                tcp_buffer.fixed.size = tcp_bs;
                globus_ftp_client_operationattr_set_tcp_buffer(oper_attr, &(tcp_buffer));
            }
            /*disable data channel authentication */
            dcau.mode = GLOBUS_FTP_CONTROL_DCAU_NONE;
            globus_ftp_client_operationattr_set_dcau(oper_attr, &(dcau));
            /* Copy FTP attributes */
            globus_gass_copy_attr_set_ftp(attr, oper_attr);
        }
            break;
        case GLOBUS_GASS_COPY_URL_MODE_GASS:
        {
            /* Get Url schema */
            globus_url_t url_struct;
            int rc = globus_url_parse(file_name, &url_struct);
            if (GLOBUS_SUCCESS != rc) {
                globus_libc_fprintf(stderr, " globus_url_parse error %d\n", rc);
            }
            /* Initialize URL transfer attributes */
            globus_gass_transfer_requestattr_init(gass_attr, url_struct.scheme);
            globus_gass_transfer_requestattr_set_file_mode(gass_attr, GLOBUS_GASS_TRANSFER_FILE_MODE_BINARY);
            /* Copy gass attributes */
            globus_gass_copy_attr_set_gass(attr, gass_attr);
        }
            break;
        default:
            break;
    } /* end switch on url_mode */

    return url_mode;
}

/*check if a file exists*/
globus_result_t dmutils_gridftp_get_exists(
        char const* file,
        char* errbuf,
        int errbufsz,
        int timeout) {

    globus_ftp_client_complete_callback_t local_complete_callback;
    globus_result_t gresult;
    monitor_t monitor;
    char *p = NULL;
    int save_errno = 0;
    local_complete_callback = &dmutils_gridftp_gcallback;
    
    if (activated == 0) {
        internal_activation = 1;     
        do_globus_activates();
    }     

    char errbuf_l[ERROR_BUFFER_SIZE];

    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }

    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.done = GLOBUS_FALSE;
    monitor.errflag = GLOBUS_FALSE;

    globus_gass_copy_url_mode_t url_mode;

    gresult = globus_gass_copy_get_url_mode(file, &url_mode);

    if (gresult != GLOBUS_SUCCESS) {
        globus_object_t *errobj;
        monitor.errflag = GLOBUS_TRUE;
        errobj = globus_error_get(gresult);
        p = GLOBUS_ERROR2STRING(errobj);
        dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
        globus_object_free(errobj);
        goto err;
    }
    if (url_mode == GLOBUS_GASS_COPY_URL_MODE_FTP) {

        /* Get File mode */
        gresult = globus_ftp_client_exists(&ftp_handle, file, &ftp_attr,
                local_complete_callback, &monitor);

        /* check the state immdiately */
        if (gresult != GLOBUS_SUCCESS) {
            save_errno = errno;
            globus_object_t *errobj;
            monitor.errflag = GLOBUS_TRUE;
            errobj = globus_error_get(gresult);
            p = GLOBUS_ERROR2STRING(errobj);
            dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
            globus_object_free(errobj);
            goto err;
        }

        int check_status = dmutils_gridftp_client_wait(&monitor, &ftp_handle, timeout);
        if (check_status != 0)
            save_errno = errno;

        if (monitor.errflag) {
            int err;
            p = GLOBUS_ERROR2STRING(monitor.error);
            globus_object_free(monitor.error);
err:
            err = scan_errstring(p);
            if (err == EEXIST)
                err = EINVAL;
            if (p != NULL) {
                dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
            }

            globus_libc_free(p);

            if (!save_errno)
                save_errno = err;
        }
    }
    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);
    
    if(internal_activation == 1)
    {
        do_globus_deactivates();
        internal_activation = 0;
    }    


    if (monitor.errflag == 0 && !save_errno) {
        errno = 0;
        return GLOBUS_TRUE;
    }

    errno = 0;
    return GLOBUS_FALSE;
}

globus_result_t check_if_exists_delete(
        char const* file,
        char* errbuf,
        int errbufsz,
        int timeout) {
    if (GLOBUS_TRUE == dmutils_gridftp_get_exists(file, errbuf, errbufsz, timeout)) /*file exists, delete it*/
        if (GLOBUS_TRUE == dmutils_gridftp_delete(file, errbuf, errbufsz, timeout)) /*attempt to delete it*/
            /*no matter if file is deleted or not, start transfering*/
            return GLOBUS_TRUE;

    return GLOBUS_TRUE;
}

globus_result_t dmutils_gridftp_delete(
        const char* file,
        char* errbuf,
        int errbufsz,
        int timeout) {

    if (activated == 0) {
        internal_activation = 1;     
        do_globus_activates();
    } 


    globus_ftp_client_complete_callback_t local_complete_callback;
    globus_result_t gresult;
    monitor_t monitor;
    char *p = NULL;
    int save_errno = 0;
    char errbuf_l[ERROR_BUFFER_SIZE];
    local_complete_callback = &dmutils_gridftp_gcallback;

 
    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }

    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.done = GLOBUS_FALSE;
    monitor.errflag = GLOBUS_FALSE;

    gresult = globus_ftp_client_delete(&ftp_handle, file, &ftp_attr,
            local_complete_callback, &monitor);

    /* check the state immdiately */
    if (gresult != GLOBUS_SUCCESS) {
        save_errno = errno;
        globus_object_t *errobj;
        monitor.errflag = GLOBUS_TRUE;
        errobj = globus_error_get(gresult);
        p = GLOBUS_ERROR2STRING(errobj);
        globus_object_free(errobj);
        goto err;
    }

    if (dmutils_gridftp_client_wait(&monitor, &ftp_handle, timeout))
        save_errno = errno;

    if (monitor.errflag) {
        int err;
        p = GLOBUS_ERROR2STRING(monitor.error);
        globus_object_free(monitor.error);
err:
        err = scan_errstring(p);
        if (err == EEXIST)
            err = EINVAL;

        if (p != NULL) {
            dmutils_gridftp_errmsg(errbuf, errbufsz, "[GRIDFTP][globus_ftp_client_delete][] %s", p);

            globus_libc_free(p);
        }
        if (!save_errno)
            save_errno = err;
    }

    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);

    if(internal_activation == 1)
    {
        do_globus_deactivates();
        internal_activation = 0;
    }

    if (monitor.errflag == 0 && !save_errno)
        errno = 0;
    return GLOBUS_TRUE;

    errno = save_errno;
    return GLOBUS_FALSE;
}

globus_result_t dmutils_gridftp_make_dir(
        char* dir,
        char *errbuf,
        int errbufsz,
        int timeout) {

    if (activated == 0) {
        internal_activation = 1;       
        do_globus_activates();
    }    

    globus_ftp_client_complete_callback_t local_complete_callback;
    globus_result_t gresult;
    monitor_t monitor;
    char *p;
    int save_errno = 0;
    char errbuf_l[ERROR_BUFFER_SIZE];
    local_complete_callback = &dmutils_gridftp_gcallback;

    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }

    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.done = GLOBUS_FALSE;
    monitor.errflag = GLOBUS_FALSE;
    
    globus_callback_register_signal_handler(
            GLOBUS_SIGNAL_INTERRUPT,
            GLOBUS_TRUE,
            globus_l_guc_interrupt_handler,
            &monitor);    

    gresult = globus_ftp_client_mkdir(&ftp_handle, dir, &ftp_attr,
            local_complete_callback, &monitor);

    /* check the state immdiately */
    if (gresult != GLOBUS_SUCCESS) {
        save_errno = errno;
        globus_object_t *errobj;
        monitor.errflag = GLOBUS_TRUE;
        errobj = globus_error_get(gresult);
        p = GLOBUS_ERROR2STRING(errobj);
        globus_object_free(errobj);
        goto err;
    }

    if (dmutils_gridftp_client_wait(&monitor, &ftp_handle, timeout))
        save_errno = errno;

    if (monitor.errflag) {
        int err;
        p = GLOBUS_ERROR2STRING(monitor.error);
        globus_object_free(monitor.error);
err:
        err = scan_errstring(p);

        if (p != NULL) {
            dmutils_gridftp_errmsg(errbuf, errbufsz, "%s", p);

            globus_libc_free(p);
        }
        if (!save_errno)
            save_errno = err;
    }


    if(!globus_l_globus_url_copy_ctrlc_handled)
    {
        globus_callback_unregister_signal_handler(
            GLOBUS_SIGNAL_INTERRUPT, GLOBUS_NULL, GLOBUS_NULL);
    }

    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);

    if(internal_activation == 1)
    {
        do_globus_deactivates();
        internal_activation = 0;
    }

    if (monitor.errflag == 0 && !save_errno)
        errno = 0;
    return GLOBUS_TRUE;

    errno = save_errno;
    return GLOBUS_FALSE;
}

globus_result_t dmutils_gridftp_get_filesize(
        char const* file,
        globus_off_t* size,
        char* errbuf,
        int errbufsz,
        int timeout) {


    if (activated == 0) {
        internal_activation = 1;
        do_globus_activates();
    }


	    
    globus_ftp_client_complete_callback_t local_complete_callback;
    globus_result_t gresult;
    monitor_t monitor;
    char *p = NULL;
    int save_errno = 0;
    local_complete_callback = &dmutils_gridftp_gcallback;

    char errbuf_l[ERROR_BUFFER_SIZE];

    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }



    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.done = GLOBUS_FALSE;
    monitor.errflag = GLOBUS_FALSE;

    globus_callback_register_signal_handler(
            GLOBUS_SIGNAL_INTERRUPT,
            GLOBUS_TRUE,
            globus_l_guc_interrupt_handler,
            &monitor);

    globus_gass_copy_url_mode_t url_mode;

    gresult = globus_gass_copy_get_url_mode(file, &url_mode);

    if (gresult != GLOBUS_SUCCESS) {
        globus_object_t *errobj;
	save_errno = errno;
        monitor.errflag = GLOBUS_TRUE;
        errobj = globus_error_get(gresult);
        p = GLOBUS_ERROR2STRING(errobj);
        dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
        globus_object_free(errobj);
        goto err;
    }
    /* switch on URL Mode */
    switch (url_mode) {
        case GLOBUS_GASS_COPY_URL_MODE_FTP:
        {
            /* Get File Size */
            gresult = globus_ftp_client_size(&ftp_handle, file, &ftp_attr,
                    size, local_complete_callback, &monitor);

            /* check the state immdiately */
            if (gresult != GLOBUS_SUCCESS) {
                save_errno = errno;
                globus_object_t *errobj;
                monitor.errflag = GLOBUS_TRUE;
                errobj = globus_error_get(gresult);
                p = GLOBUS_ERROR2STRING(errobj);
                dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
                globus_object_free(errobj);
                goto err;
            }

            int check_status = dmutils_gridftp_client_wait(&monitor, &ftp_handle, timeout);
            if (check_status != 0)
                save_errno = errno;

            if (monitor.errflag) {
                int err;
                p = GLOBUS_ERROR2STRING(monitor.error);
                globus_object_free(monitor.error);
err:
                err = scan_errstring(p);
                if (err == EEXIST)
                    err = EINVAL;
                if (p != NULL) {
                    dmutils_gridftp_errmsg(errbuf, errbufsz, " %s", p);
                }

                globus_libc_free(p);

                if (!save_errno)
                    save_errno = err;
            }
        }
            break;
        case GLOBUS_GASS_COPY_URL_MODE_IO:
        {
            /* Local file */
            char* local_name = file;
            if (strncmp(file, "file://", 7) == 0) {
                substr(local_name, file, 7, strlen(file));
            }

            struct stat64 stat_buf;
            int r = stat64(local_name, &stat_buf);
            if (0 != r) {
                save_errno = errno;
            }
            /* Set File Size */
            *size = stat_buf.st_size;
        }
            break;
        case GLOBUS_GASS_COPY_URL_MODE_GASS:
            globus_libc_fprintf(stderr, " file type is GLOBUS_GASS_COPY_URL_MODE_GASS \n");
            break;
        default:
            globus_libc_fprintf(stderr, " Unsupported file url type \n");
            break;
    }

 if(!globus_l_globus_url_copy_ctrlc_handled)
    globus_callback_unregister_signal_handler(GLOBUS_SIGNAL_INTERRUPT, GLOBUS_NULL, GLOBUS_NULL);
    
    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);



    if(internal_activation == 1)
    {
    	do_globus_deactivates();
        internal_activation = 0;  
    }

    if (monitor.errflag == 0 && !save_errno) {
        errno = 0;
        return GLOBUS_TRUE;
    }

    errno = save_errno;
    return GLOBUS_FALSE;
}

globus_result_t dmutils_gridftp_checksum(
        char const* file,
        char* cksm,
        char const* cksm_alg,
        char* errbuf,
        int errbufsz,
        int timeout) {


    if (activated == 0) {
        internal_activation = 1;
        do_globus_activates();
    }   

    char checksum[256] = {0};

    globus_ftp_client_complete_callback_t local_complete_callback;
    globus_result_t gresult;
    monitor_t monitor;
    char *p = NULL;
    int save_errno = 0;
    local_complete_callback = &dmutils_gridftp_gcallback;

    char errbuf_l[ERROR_BUFFER_SIZE];
    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }

    if (file == NULL || cksm == NULL) {
        dmutils_gridftp_errmsg(errbuf, errbufsz, "[gridftp_checksum][EINVAL] Invalid arguments");
        errno = EINVAL;
        return GLOBUS_FALSE;
    }
    memset((void*) checksum, 0, 256);

    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);
    monitor.done = GLOBUS_FALSE;
    monitor.errflag = GLOBUS_FALSE;

    globus_callback_register_signal_handler(
            GLOBUS_SIGNAL_INTERRUPT,
            GLOBUS_TRUE,
            globus_l_guc_interrupt_handler,
            &monitor);
    
    gresult = globus_ftp_client_cksm(
            &ftp_handle,
            file,
            &ftp_attr,
            checksum,
            (globus_off_t) 0,
            (globus_off_t) - 1,
            cksm_alg,
            local_complete_callback,
            &monitor);

    /* check the state immdiately */
    if (gresult != GLOBUS_SUCCESS) {
        save_errno = errno;
        globus_object_t *errobj;
        monitor.errflag = GLOBUS_TRUE;
        errobj = globus_error_get(gresult);
        p = GLOBUS_ERROR2STRING(errobj);
        dmutils_gridftp_errmsg(errbuf, errbufsz, p, file);
        globus_object_free(errobj);
        goto err;
    }

    int check_status = dmutils_gridftp_client_wait(&monitor, &ftp_handle, timeout);
    if (check_status != 0)
        save_errno = errno;


    if (monitor.errflag) {
        int err;
        p = GLOBUS_ERROR2STRING(monitor.error);
        globus_object_free(monitor.error);
err:
        err = scan_errstring(p);
        dmutils_gridftp_errmsg(errbuf, errbufsz, "%s: CKSM (checksum) operation error", p);
        if (p) globus_libc_free(p);

        if (!save_errno) {
            save_errno = err;
            if (err == EOPNOTSUPP)
                dmutils_gridftp_errmsg(errbuf, errbufsz, "%s: CKSM (checksum) operation not supported", file);
        }
    }
	if(!globus_l_globus_url_copy_ctrlc_handled)
    		globus_callback_unregister_signal_handler(GLOBUS_SIGNAL_INTERRUPT, GLOBUS_NULL, GLOBUS_NULL);

    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);

    if(internal_activation == 1)
    {
        do_globus_deactivates();
        internal_activation = 0;
    }    

    if (monitor.errflag == 0 && !save_errno) {
        strncpy(cksm, checksum, strlen(checksum) + 1);
        errno = 0;
        return GLOBUS_TRUE;
    }

    dmutils_gridftp_errmsg(errbuf, errbufsz, "%s: CKSM (checksum) operation aborted", file);
    errno = save_errno;
    return GLOBUS_FALSE;
}

/**
 * Gass copy transfer performance callback
 *
 * This callback is registered with 'globus_gass_copy_register_performance_cb'
 * It will be called during a transfer to supply performance information on
 * current transfer.  Its frequency will be at most one per second, but
 * it is possible to receive no callbacks. This is possible in very short
 * transfers and in ftp transfers in which the server does not provide
 * performance information.
 *
 */
void dmutils_gridftp_copy_performance(
        void * user_arg,
        globus_gass_copy_handle_t * handle,
        globus_off_t total_bytes,
        float instantaneous_throughput,
        float avg_throughput) {
    s_transfer_performance.transferred_bytes = total_bytes;
    s_transfer_performance.instant_throughput = instantaneous_throughput;
    s_transfer_performance.avg_throughput = avg_throughput;
    usercb_ts.tv_sec = time(0);
    usercb_ts.tv_nsec = 0;

}

void copy_print_callback(performance perf, struct timespec ts, int *timeoutflg) {

    fprintf(stderr, "\r %12ld bytes", (LONG64) perf.transferred_bytes);
    fprintf(stderr, " %9.2lf KB/sec avg", (double) perf.avg_throughput / 1024);
    fprintf(stderr, " %9.2lf KB/sec inst\n", (double) perf.instant_throughput / 1024);

}

void copy_print_callback_unused(performance perf, struct timespec ts, int *timeoutflg) {
    /*no output or action is required here*/
}

void dmutils_gridftp_copy_callback(
        void *callback_arg,
        globus_gass_copy_handle_t *handle,
        globus_object_t *error) {
    monitor_t *monitor = (monitor_t *) callback_arg;
    
    if(monitor){
    globus_mutex_lock(&monitor->mutex);
    if (error != GLOBUS_SUCCESS) {
        monitor->errflag = GLOBUS_TRUE;
        monitor->error = globus_object_copy(error);
    }
    monitor->done = GLOBUS_TRUE;
    globus_cond_signal(&monitor->cond);
    globus_mutex_unlock(&monitor->mutex);
    }
}

/**
 * Transfer data from source URL to destination URL
 *
 * This functions initiates a transfer from source URL to destination URL,
 * then returns immediately.
 *
 * When the transfer is completed or if the
 * transfer is aborted, the callback_func will be invoked with the final
 * status of the transfer.
 *
 */
globus_result_t dmutils_gridftp_gass_copy_url_to_url(
        const char* src_file,
        const char* dest_file,
        int const nbstreams,
        int const tcpbs,
        int const block_size,
        char* errbuf,
        int errbufsz,
        int const timeout,
        int verbose,
        globus_gass_copy_handle_t *ggc_handle_external,
        void (*ggc_callback)(void *callback_arg, globus_gass_copy_handle_t *handle, globus_object_t *error),
        void * monitor_external,
        void (*globus_gass_copy_performance_cb_t_etx)(void *user_arg,
        globus_gass_copy_handle_t *handle,
        globus_off_t total_bytes,
        float instantaneous_throughput,
        float avg_throughput),
        void * check_external,
        int (*gridftp_copy_wait)(void * monitor_wait, void * check_wait, globus_gass_copy_handle_t *gfhp_wait, int const timeout_wait, void *iupdater_wait),
        void * iupdater) {
    /*if the last 6 parameters are NULL, then a transfer with default callbacks will be used*/

    /*declare function pointer callback for external copy waiting*/
    int (*gridftp_copy_wait_local)(void * monitor, void * check, globus_gass_copy_handle_t *gfhp, int const timeout, void * iupdater) = NULL;

    copyfile_callback_t cp_callback = NULL;

    globus_result_t gresult;
    monitor_t monitor;
    char *p = NULL;
    struct timespec ts;
    check_t check;
    int save_errno = 0;

    istimeout = 0;
    s_transfer_performance.transferred_bytes = 0;


    globus_mutex_init(&monitor.mutex, NULL);
    globus_cond_init(&monitor.cond, NULL);

    char errbuf_l[ERROR_BUFFER_SIZE];
    if (errbuf == NULL || errbufsz == 0) {
        errbufsz = ERROR_BUFFER_SIZE;
        errbuf = errbuf_l;
    }

    /* set block size option */
    if (block_size > 0) {
        gresult = globus_gass_copy_set_buffer_length(&ggc_handle, block_size);
        if (GLOBUS_SUCCESS != gresult) {
            monitor.errflag = GLOBUS_TRUE;
            globus_object_t * obj = globus_error_get(gresult);
            errbuf = dmutils_gridftp_error_to_string(obj);
            globus_object_free(obj);
        }
    }

    globus_gass_copy_set_allocate(&ggc_handle, GLOBUS_TRUE);

    /* Prepare Attributes */
    globus_gass_copy_url_mode_t src_url_mode;
    globus_gass_copy_url_mode_t dst_url_mode;

    /* Prepare source file attributes */
    src_url_mode = prepare_file_attr(src_file, "source", nbstreams, tcpbs, &source_gass_copy_attr, &source_ftp_attr, &source_gass_attr);
    /* Prepare destination file attributes */
    dst_url_mode = prepare_file_attr(dest_file, "dest", nbstreams, tcpbs, &dest_gass_copy_attr, &dest_ftp_attr, &dest_gass_attr);


    if (src_url_mode == -1 || dst_url_mode == -1) {
        monitor.done = GLOBUS_TRUE;
        monitor.errflag = GLOBUS_TRUE;
        dmutils_gridftp_errmsg(errbuf, errbufsz, "Fail to prepare file attr for source or destination file");
        return GLOBUS_FALSE;
    }


    if (check_external == NULL) {

        globus_callback_register_signal_handler(
                GLOBUS_SIGNAL_INTERRUPT,
                GLOBUS_TRUE,
                globus_l_guc_interrupt_handler,
                &monitor);          

        /* Register performance callback */
        if (verbose) {
            cp_callback = &copy_print_callback;
        } else {
            cp_callback = &copy_print_callback_unused;
        }

        monitor.done = GLOBUS_FALSE;
        monitor.errflag = GLOBUS_FALSE;


        globus_gass_copy_register_performance_cb(&ggc_handle, dmutils_gridftp_copy_performance, GLOBUS_NULL);

        gresult = globus_gass_copy_register_url_to_url(&ggc_handle,
                src_file, &source_gass_copy_attr, dest_file, &dest_gass_copy_attr,
                &dmutils_gridftp_copy_callback, &monitor);

        /* check the state immdiately */
        if (gresult != GLOBUS_SUCCESS) {
            globus_object_t *errobj;
            monitor.errflag = GLOBUS_TRUE;
            errobj = globus_error_get(gresult);
            p = GLOBUS_ERROR2STRING(errobj);
            globus_object_free(errobj);
            goto err;
        }

        globus_mutex_lock(&monitor.mutex);
        check.user_cb = &cp_callback;
        check.mt = &monitor;
        check.done = GLOBUS_FALSE;
        globus_callback_register_oneshot(
                NULL,
                NULL,
                dmutils_gridftp_timeout_callback,
                &check);
        if (timeout) {
            ts.tv_sec = time(0) + timeout;
            ts.tv_nsec = 0;
            while ((!monitor.done) && (!istimeout) && (!save_errno)) {
                if (globus_l_globus_url_copy_ctrlc &&
                        !globus_l_globus_url_copy_ctrlc_handled) {
                    globus_gass_copy_cancel(&ggc_handle,
                            &dmutils_gridftp_copy_callback, &monitor);
                    globus_l_globus_url_copy_ctrlc_handled = GLOBUS_TRUE;
                }
                save_errno = globus_cond_timedwait(&monitor.cond, &monitor.mutex, &ts);
                if (save_errno == EINTR)
                    save_errno = 0;
            }
        } else
            while ((!monitor.done) && (!istimeout) && (!save_errno)) {
                if (globus_l_globus_url_copy_ctrlc &&
                        !globus_l_globus_url_copy_ctrlc_handled) {
                    globus_gass_copy_cancel(&ggc_handle,
                            &dmutils_gridftp_copy_callback, &monitor);
                    globus_l_globus_url_copy_ctrlc_handled = GLOBUS_TRUE;
                }
                save_errno = globus_cond_wait(&monitor.cond, &monitor.mutex);
                if (save_errno == EINTR)
                    save_errno = 0;
            }
        if (save_errno == ETIMEDOUT) {
            istimeout = 1;
        }

        if (!monitor.done) {
            int ret;
            globus_gass_copy_cancel(&ggc_handle, &dmutils_gridftp_copy_callback, &monitor);
            /* wait for the cancel to complete before returning to user */
            do {
	    	struct timespec ts;
                ts.tv_sec = time(0) + 60;
                ts.tv_nsec = 0;
                ret = globus_cond_timedwait(&monitor.cond, &monitor.mutex, &ts);
                if (ret == EINTR)
                    ret = 0;
            } while (!monitor.done && !ret);
            if (ret && !save_errno)
                save_errno = ret;
        }
        /* wait timeout_callback to complete */
        if (!check.done) {
            int ret;
            do {
	    	struct timespec ts;
                ts.tv_sec = time(0) + 60;
                ts.tv_nsec = 0;
                ret = globus_cond_timedwait(&monitor.cond, &monitor.mutex, &ts);
                if (ret == EINTR)
                    ret = 0;
            } while (!check.done && !ret);
            if (ret && !save_errno)
                save_errno = ret;
        }
        globus_mutex_unlock(&monitor.mutex);

        if (globus_l_globus_url_copy_ctrlc_handled == GLOBUS_TRUE) {
            save_errno = ECANCELED;
        } else if (istimeout == 1) {
            save_errno = ETIMEDOUT;
        }

        if (monitor.errflag) {
            int err;
            p = GLOBUS_ERROR2STRING(monitor.error);
            globus_object_free(monitor.error);
err:

            err = scan_errstring(p);
            if (p != NULL && !save_errno)
                dmutils_gridftp_errmsg(errbuf, errbufsz, "%s", p);
            if (p != NULL)
                globus_libc_free(p);

            if (!save_errno)
                save_errno = err;
        }
     if(!globus_l_globus_url_copy_ctrlc_handled)
	globus_callback_unregister_signal_handler(GLOBUS_SIGNAL_INTERRUPT, GLOBUS_NULL, GLOBUS_NULL);

        globus_mutex_destroy(&monitor.mutex);
        globus_cond_destroy(&monitor.cond);

        if ((monitor.errflag == 0) && (!save_errno)) {
            errno = 0;
            return GLOBUS_TRUE;
        }

        errno = save_errno;
        return GLOBUS_FALSE;
    } else {
    
        globus_gass_copy_register_performance_cb(&ggc_handle, globus_gass_copy_performance_cb_t_etx, check_external);

        gresult = globus_gass_copy_register_url_to_url(&ggc_handle,
                src_file, &source_gass_copy_attr, dest_file, &dest_gass_copy_attr,
                ggc_callback, monitor_external);
        if (GLOBUS_SUCCESS != gresult) {
            globus_object_t *errobj;
            errobj = globus_error_get(gresult);
            p = GLOBUS_ERROR2STRING(errobj);
            globus_object_free(errobj);
            if (p != NULL)
                dmutils_gridftp_errmsg(errbuf, errbufsz, "%s", p);
            if (p != NULL)
                globus_libc_free(p);
            goto copy_error;
        }

        gridftp_copy_wait_local = gridftp_copy_wait;
        int result = gridftp_copy_wait_local(monitor_external, check_external, &ggc_handle, timeout, iupdater);

        if (0 != result) {
            errno = result;
            goto copy_error;
        }

copy_error:
       
        if ((GLOBUS_SUCCESS != gresult) || (0 != result)){
    	    errno = result;
            return GLOBUS_FALSE;
	    }

        errno = 0;
        return GLOBUS_TRUE;
    }
}

/*
 * Provide a generic way of transfering files: with or without parameters
 * This function can cover the whole spectrum of transfers(put, get, TPT)
 */
globus_result_t dmutils_gridftp_copy(
        const char* source_url,
        const char* dest_url,
        char *errbuf,
        int errbufsz,
        int timeout,
        int const nbstreams,
        int const tcp_bs,
        int const block_size,
        int verbose,
        globus_gass_copy_handle_t *ggc_handle,
        void (*ggc_callback)(void *callback_arg, globus_gass_copy_handle_t *handle, globus_object_t *error),
        void * monitor_external,
        void (*globus_gass_copy_performance_cb_t_etx)(void *user_arg,
        globus_gass_copy_handle_t *handle,
        globus_off_t total_bytes,
        float instantaneous_throughput,
        float avg_throughput),
        void * check_external,
        int (*gridftp_copy_wait)(void * monitor, void * check, globus_gass_copy_handle_t *gfhp, int const timeout, void * iupdater),
        void * iupdater) {

      if (activated == 0) {
        internal_activation = 1;
        do_globus_activates();
    }

    globus_result_t result =  dmutils_gridftp_gass_copy_url_to_url(
            source_url,
            dest_url,
            nbstreams,
            tcp_bs,
            block_size,
            errbuf,
            errbufsz,
            timeout,
            verbose,
            ggc_handle,
            ggc_callback,
            monitor_external,
            globus_gass_copy_performance_cb_t_etx,
            check_external,
            gridftp_copy_wait,
            iupdater);

    if(internal_activation == 1)
    {
        do_globus_deactivates();
        internal_activation = 0;
    }

   return result;
}

char* dmutils_gridftp_error_to_string(
        globus_object_t* errobj) {
    char* msg = 0;

    if (0 != errobj) {
#ifdef GLOBUS_ERROR_TYPE_MULTIPLE
        char * p = globus_error_print_friendly(errobj);
#else
        char * p = globus_object_printable_to_string(errobj);
#endif
        if (0 != p) {
            msg = p;
            globus_libc_free(p);
            return msg;
        }
    }
    return 0;
}

#endif
